using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Text;

using CSPM.Client.Com;
using CSPM.Client.Obj;

namespace CSPM.Client.Data
{
    public sealed class ConnectionFactory  
    {
        // Constants
        private const string MSG_CSPM_ERROR = 
    	    "Error retrieving credentials. Error: {0} - {1}";

        private const string STR_UID = "uid=";
        private const string STR_USER_ID = "User ID=";
        
        private const string STR_PWD = "pwd=";
        private const string STR_PASSWORD = "password=";
        
        private const string STR_ALIAS = "CSPMAlias=";

        private const string ERR_MISSING_UID = 
            "Connection string missing \"user id\" or \"uid\" element";
        private const string ERR_MISSING_PWD = 
            "Connection string missing \"password\" or \"pwd\" element";

        /*
         * Make the default constructor private.
         */
        private ConnectionFactory()
        {
        }

        /*
         * Create and Open SqlConnection object. 
         * 
         * @param connectionString
         * @param alias
         * @return SqlConnection
         */
        static public SqlConnection OpenSqlConnection(
                string connectionString, string alias) 
        {
            // Make a copy of the alias
            string aliasName = alias;

            // StringBuilder used to modify the connection string
            StringBuilder connectStr = new StringBuilder(connectionString);

            // CSPM Client object used to retrieve the credentials.
            CspmClientObject cspmClient = null;

            // SqlConnection to return.
            SqlConnection connection = null;

            try
            {
                // if the connection string contains an alias, extract if.
                extractAliasName(ref connectStr, ref aliasName);

                // if we have an alias.
                if (aliasName != null)
                {
                    // Create CSPMClient COM Object 
                    cspmClient = new CspmClientComObject();

                    // Retrieve the credentials
                    retrieveCredentials(cspmClient, aliasName, "false");

                    // Update the connection string with the uid and pwd
                    updateCredentials(ref connectStr, cspmClient);
                }

                // Create the SqlConnection object
                connection = new SqlConnection(connectStr.ToString());

                try
                {
                    // Try open it to verify a possible invalid 
                    // credentials issue.
                    connection.Open();
                }
                catch (SqlException sqlEx)
                {
                    // if not authentication error.
                    if (sqlEx.Number != 0x4818)
                    {
                        // Rethrow.
                        throw;
                    }

                    // Save the current information
                    string oldUserId = cspmClient.GetUserId;
                    string oldPassword = cspmClient.GetPassword;

                    // Retrieve the credentials from the server.
                    retrieveCredentials(cspmClient, aliasName, "true");

                    // if retrieved the same information
                    if (oldUserId.Equals(cspmClient.GetUserId) &&
                            oldPassword.Equals(cspmClient.GetPassword))
                    {
                        // Rethrow the same exception
                        throw;
                    }

                    // Update the connection string with the uid and pwd
                    updateCredentials(ref connectStr, cspmClient);

                    // Reset the connection string.
                    connection.ConnectionString = connectStr.ToString();

                    // Try again.
                    connection.Open();
                }
            }
            finally
            {
                // if need to dispose of the CSPMClienComObject
                if (cspmClient != null)
                {
                    // Dispose of the CSPM Client object.
                    cspmClient.Dispose();
                }
            }

            // Return the newly created connection.
            return connection;
        }

        /*
         * Extract the alias name from the connection string. 
         * 
         * @param connectStr
         * @param aliasName
         * 
         */
        static private void extractAliasName(ref StringBuilder connectStr, 
                ref string aliasName)
        {
            // if the connection string contains an alias.
            int pos = connectStr.ToString().IndexOf(STR_ALIAS, 
                    StringComparison.CurrentCultureIgnoreCase);
            if (pos >= 0)
            {
                // Starting position of the alias name.
                int aliasPos = pos + STR_ALIAS.Length;

                // Find the end of the element
                int end = connectStr.ToString().IndexOf(';', aliasPos);

                // if not ending with a semicolon
                if (end == -1)
                {
                    // Use the length of the string.
                    end = connectStr.Length;
                }

                // If need to extract the name
                if (aliasName == null)
                {
                    // Extract the name
                    aliasName = connectStr.ToString().Substring(
                        aliasPos, end - aliasPos);
                }

                // Remove the alias from the string
                connectStr.Remove(pos, end - pos);
            }
        }

        /*
         * Refresh the uid and password elements in the connection string.
         * 
         * @param connectStr
         * @param cspmClient
         */
        static private void updateCredentials(ref StringBuilder connectStr, 
                CspmClientObject cspmClient)
        {
            // Try replacing the uid elt
            if (!replaceElement(ref connectStr, STR_UID, 
                    cspmClient.GetUserId))
            {
                // Try replacing the User ID elt
                if (!replaceElement(ref connectStr, STR_USER_ID, 
                        cspmClient.GetUserId))
                {
                    // Report an error
                    throw new InvalidOperationException(ERR_MISSING_UID);
                }
            }

            // Try replacing the pwd elt
            if (!replaceElement(ref connectStr, STR_PWD, 
                    cspmClient.GetPassword))
            {
                // Try replacing the password elt
                if (!replaceElement(ref connectStr, STR_PASSWORD, 
                        cspmClient.GetPassword))
                {
                    // Report an error
                    throw new InvalidOperationException(ERR_MISSING_PWD);
                }
            }
        }

        /*
         * Locate and replace and element value in the connection string.
         * Return true if the element was replaced.
         * 
         * @param ref connectStr
         * @param toFind
         * @param toReplace
         * @return bool 
         */
        static private bool replaceElement(ref StringBuilder connectStr, 
                string toFind, string toReplace)
        {
            // if the connection string contains the element.
            int pos = connectStr.ToString().IndexOf(toFind, 
                StringComparison.CurrentCultureIgnoreCase);

            // if the element is missing
            if (pos == -1)
            {
                return false;
            }

            // Move pos to the beginning of the element value.
            pos += toFind.Length;
            
            // Locate the end of it
            int end = connectStr.ToString().IndexOf(';', pos);

            // if not ending with a semicolon
            if (end == -1)
            {
                // Use the end.
                end = connectStr.Length;
            }

            // Replace the value
            connectStr.Remove(pos, end - pos);
            connectStr.Insert(pos, toReplace);

            return true;
        }

        /*
         * Retrieve the credentials using the CSPMClientObject.
         * Throws an InvalidOperationException if not a valid call.
         * 
         * @param cspmClient
         * @param alias
         * @param how  String containing "true" or "false".
         * 
         */
        static private void retrieveCredentials(CspmClientObject cspmClient, 
                string alias, string how)
        {
            // Retrieve the credentials
            Int32 statusCode = cspmClient.RetrieveCredentials(
                alias, how, "");
            if (statusCode != 400)
            {
                StringBuilder errMsg = new StringBuilder();
                errMsg.AppendFormat(MSG_CSPM_ERROR,
                    statusCode.ToString("d"),
                    cspmClient.GetStatusMsg(statusCode));
                throw new InvalidOperationException(errMsg.ToString());
            }
        }
    }
}